from typing import Union, Optional, Any, NamedTuple, Literal
import math
import re

Style = Literal['kanji', 'arabic', 'mix', 'finance', 'daiji']

# Grok for Japanese Number
NUMBERS = (
    r'('
      r'[01-9]+'
    r')'
)

NOT_NUMBERS = (
    r'([^01-9]+)?'
)

KANJI_DIGITS = (
    r'〇一二三四五六七八九'
    r'０１２３４５６７８９'
)
KANJI_ARABIC_DIGITS = (
    r'０１２３４５６７８９'
)

KANJI_UNITS = (
    r'十百千万億兆京垓𥝱'
)

# Old style numbers.
DAIJI_DIGITS = (
    r'零壱弐参肆伍陸漆捌玖'
    r'拾佰仟萬'
)

YEN_SYMBOL = (
    r'[円圓]'
)

NOT_KANJI_NUMBERS = (
    r'([^'
       f'{KANJI_DIGITS}十'
    r']*)?'
)

KANJI_NUMBERS = (
    r'(['
      f'{KANJI_DIGITS}十'
    r']+)'
)

KANJI_ARABIC_NUMBERS = (
    r'(['
      f'{KANJI_ARABIC_DIGITS}'
    r']*)'
)

NOT_KANJI_ARABIC_NUMBERS = (
    r'([^'
      f'{KANJI_ARABIC_DIGITS}'
    r']*)'
)

class JpNumber(NamedTuple):
    number: Optional[int]=None
    as_str: Optional[str]=None
    as_kanji: Optional[str]=None

class ParseNumber(NamedTuple):
    prefix: str
    number: JpNumber
    suffix: str

class Token(NamedTuple):
    as_int: int
    as_str: str
    unit: int

class JpNumberParser(object):

    __KANJI_DIGITS=(
        '〇', '一', '二', '三', '四', '五', '六', '七', '八', '九')
    __MULTIBYTE_DIGITS=(
        '０', '１', '２', '３', '４', '５', '６', '７', '８', '９')
    __DAIJI_DIGITS=(
        '零', '壱', '弐', '参', '肆', '伍', '陸', '漆', '捌', '玖')

    __KANJI_MINOR_UNITS = {
        '千': 1000,
        '百': 100,
        '十': 10,
    }
    __KANJI_MAJOR_UNITS = ( '𥝱', '垓', '兆', '京', '億', '万' )

    __KANJI_UNIT_MAPPER = {
        '万':   pow(10, 4),
        '十万': pow(10, 5),
        '百万': pow(10, 6),
        '千万': pow(10, 7),
        '億':   pow(10, 8),
        '十億': pow(10, 9),
        '百億': pow(10, 10),
        '千億': pow(10, 11),
        '兆':   pow(10, 12),
        '十兆': pow(10, 13),
        '百兆': pow(10, 14),
        '千兆': pow(10, 15),
        '京':   pow(10, 16),
        '十京': pow(10, 17),
        '百京': pow(10, 18),
        '千京': pow(10, 19),
        '垓':   pow(10, 20),
        '十垓': pow(10, 21),
        '百垓': pow(10, 22),
        '千垓': pow(10, 23),
        '𥝱':   pow(10, 24),
        '十𥝱': pow(10, 25),
        '百𥝱': pow(10, 26),
        '千𥝱': pow(10, 27),
    }

    def __init__(self):
        self.__NUMERICS = list(map(str, range(0, 10)))
        self.__Number2KanjiNumber = {
            **{ num: kanji
                for num, kanji in enumerate(self.__KANJI_DIGITS)},
            **{ str(num): kanji
                for num, kanji in enumerate(self.__KANJI_DIGITS)}
            }
        self.__Number2ArabicNumber = {
            **{ num: kanji
                for num, kanji in enumerate(self.__MULTIBYTE_DIGITS)},
            **{ str(num): kanji
                for num, kanji in enumerate(self.__MULTIBYTE_DIGITS)}
            }
        self.__KanjiNumber2Number = {
            **{ kanji: num
                for num, kanji in enumerate(self.__KANJI_DIGITS)},
            **{ kanji: num
                 for num, kanji in enumerate(self.__MULTIBYTE_DIGITS)}
            }
        self.__unit_mapper = {
            **{ unit: kanji_unit
                for kanji_unit, unit in self.__KANJI_MINOR_UNITS.items()},
            **{ unit: kanji_unit
                for kanji_unit, unit in self.__KANJI_UNIT_MAPPER.items()},
            **{ kanji_unit: unit
                for kanji_unit, unit in self.__KANJI_MINOR_UNITS.items()},
            **{ kanji_unit: unit
                for kanji_unit, unit in self.__KANJI_UNIT_MAPPER.items()},
            }

    def _number_tokenizer(self,
            as_str
        )->list:
        size = len(as_str)
        token_list = list()
        if size == 1:
              if as_str[0] in self.__Number2KanjiNumber:
                  c = self.__Number2KanjiNumber[as_str[0]]
              token_list.append(Token(int(as_str[0]), as_str[0], -1))
              return token_list

        unit = pow(10, size)
        for c in as_str:
            if c in self.__Number2KanjiNumber:
                unit = int(unit / 10)
                if unit not in self.__unit_mapper:
                    unit = 1
                token_list.append(Token(int(c), c, unit))
        return token_list

    def _number2kanji_parser(self,
            token_list: list,
            style: Style
        )->list:
        stack = list()
        if style not in ['kanji', 'daiji']:
            return stack

        base_unit = 1
        for token in token_list:
            as_int = token.as_int
            unit = token.unit
            if as_int != 0:
                stack.append(self.__Number2KanjiNumber[as_int])
            base_unit *= 10
            if unit in self.__unit_mapper:
                kanji_unit = self.__unit_mapper[unit]
            elif unit < 0:
                if as_int == 0:
                    stack.append(self.__Number2KanjiNumber[as_int])
                break
            else:
                kanji_unit = ""
            if len(kanji_unit) >= 1:
                kanji_unit = kanji_unit[0]
            if as_int != 0:
                if kanji_unit in self.__KANJI_MINOR_UNITS:
                    stack.append( kanji_unit )
                count = 0
            else:
                count += 1
            if ( count <4 and kanji_unit in self.__KANJI_UNIT_MAPPER):
                    stack.append( kanji_unit )
        return stack

    def _number2mix_parser(self,
            token_list: list,
            style: Style
        )->list:

        count = 0
        stack = list()
        if style not in ['mix']:
            return stack

        for token in reversed(token_list):
            as_int = token.as_int
            unit = token.unit
            v = self.__Number2ArabicNumber[as_int]

            if unit in self.__unit_mapper:
                kanji_unit = self.__unit_mapper[unit]
            else:
                kanji_unit = '一'

            if len(kanji_unit) != 1:
                kanji_unit = kanji_unit[0]

            if kanji_unit in self.__KANJI_MAJOR_UNITS:
                if count < 4:
                    stack_size = len(stack) -1
                    stack[stack_size] = kanji_unit
                else:
                    stack.append(kanji_unit)
                count = 0

            stack.append(v)
            if as_int != 0:
                count = 0
            else:
                count += 1
            if count >= 4:
                del(stack[-5:])



        return reversed(stack)

    def _number2arabic_parser(self,
            token_list: list,
            style: Style
        )->list:

        stack = list()
        if style not in ['arabic', 'finance']:
            return stack

        for token in token_list:
            c = token.as_int
            v = self.__Number2ArabicNumber[c]
            stack.append(v)

        if style == 'finance':
            new_stack=list()
            sub_unit = 1
            for c in reversed(stack):
                new_stack.append(c)
                sub_unit *= 10
                if sub_unit % 1000 == 0:
                    new_stack.append('，')
                    sub_unit = 1
            if new_stack[-1] == '，':
                new_stack.pop(-1)
            stack = reversed(new_stack)
        return stack

    def number2kanji(self,
            val: Union[int, str],
            style: Style='kanji',
            trim: bool=True,
        )-> JpNumber:
        """ Conver Number to Kanji Number
        Parameters
        ----------
        val: Union[str, int]
            any positive number
        syle: Style
             output style.  default is ``kanji``
               'kanji': '一万二千三百'
              'arabic': '１２３００'
                 'mix': '１万２３００'
             'finance': '１２，３００'

        trim: bool
            if set ``True``, drop '一' from '一百' or '一千'.
            default is ``True``

        Returns:
           converted number: JpNumber
        """

        replace_chars = {
            '^一千': '千',
            '^一百': '百',
            '^一十': '十',
            '^１千': '千',
            '^１百': '百',
            '^１十': '十',
        }

        replace_daiji = {
            '〇': '零', '一': '壱',
            '二': '弐',
            '三': '参',
            '四': '肆',
            '五': '伍',
            '六': '陸',
            '七': '漆',
            '八': '捌',
            '九': '玖',
            '十': '拾',
            '百': '佰',
            '千': '仟',
            '万': '萬',
        }

        action = {
            'arabic': self._number2arabic_parser,
            'finance': self._number2arabic_parser,
            'kanji': self._number2kanji_parser,
            'daiji': self._number2kanji_parser,
            'mix': self._number2mix_parser,
        }

        if isinstance(val, int):
            number = val
            as_str = str(val)
        else:
            number = int(val)
            as_str = val

        if number <0:
            return JpNumber( number, as_str, '')

        token_list = self._number_tokenizer(as_str)
        stack = action[style](token_list, style)
        as_kanji = str().join(stack)

        if style == 'daiji':
            for src, dst in replace_daiji.items():
                as_kanji = re.sub(src, dst, as_kanji)
        elif trim:
            for src, dst in replace_chars.items():
                as_kanji = re.sub(src, dst, as_kanji)
        return JpNumber( number, as_str, as_kanji)

    def kanji2arabic_parser(self,
           val : str,
        )-> list:
        # １２３０００００
        unit = 1
        number = 0
        for c in reversed(val):
            if c in self.__KanjiNumber2Number:
                number += self.__KanjiNumber2Number[c] * unit
                unit *= 10
        return number

    def kanji2mix_parser(self,
           val : str,
        )-> list:

        # val １億２千２３０万
        #  conver strings from kanji arabic char to numeric digits.
        #  stack [1, 100000000, 2, 1000, 230, 10000]
        # calc
        #  stack[1 x 100000000, (2 x 1000 + 230) x 10000]
        # sum(stack)

        stack = list()
        as_str = ''
        for c in val:
            if c in self.__KanjiNumber2Number:
                as_str += str(self.__KanjiNumber2Number[c])
                continue
            elif c in self.__KANJI_MAJOR_UNITS:
                stack.append(int(as_str))
                stack.append(self.__unit_mapper[c])
                as_str=''
            elif c in self.__KANJI_MINOR_UNITS.keys():
                stack.append(int(as_str))
                stack.append(self.__KANJI_MINOR_UNITS[c])
                as_str=''

        number = 0
        base_unit = unit = 1
        for elm in reversed(stack):
            if elm in self.__KANJI_UNIT_MAPPER.values():
                base_unit = elm
                unit = 1
                continue
            elif elm in self.__KANJI_MINOR_UNITS.values():
                unit = elm
                continue
            else:
                number += elm * unit * base_unit
                unit = 1
        return number

    def kanji2number_parser(self,
           val : str,
        )-> list:
        """
        Parameters
        ----------
        val: str
             any positive integer as string

        Logic
        -----
            val 一億二百三十万
            Step1:  Split Major Unit
                 stack[一, 100000000, 二百三十, 10000]
            Step2:  Split Minor Unit
                 stack[一, 100000000, 二, 100, 三十, 10000]
                 stack[一, 100000000, 二, 100, 三, 10, 10000]
            Step3:   numeric strings to interger
                 stack[1 , 10000000, 2, 100, 3, 10, 10000]
            Step4:  calculartion
                 1 x 100000000 + (2 x100 + 3 x 30) x 10000
        """
        as_str = val
        number = 0
        stack = list()
        for sp in self.__KANJI_MAJOR_UNITS:
            r = as_str.split(sp)
            if len(r) != 1:
                if r[0] == '':
                    stack.append('一')
                else:
                    stack.append(r[0])
                if r[1] == '':
                    stack.append(self.__unit_mapper[sp])
                    break
                else:
                    as_str = r[1]
                    stack.append(self.__unit_mapper[sp])
            elif r[0] != '':
                as_str = r[0]
        else:
           stack.append(as_str)

        stack2 = list()
        for word in reversed(stack):
            if isinstance(word, int):
                 stack2.append(word)
                 continue

            if word in self.__KanjiNumber2Number:
                stack2.append(word)
                continue

            as_str = word
            for sp in self.__KANJI_MINOR_UNITS.keys():
                r = as_str.split(sp)
                if len(r) != 1:
                    stack2.append(self.__KANJI_MINOR_UNITS[sp])
                    if r[0] == '':
                        stack2.append('一')
                    else:
                        if r[0] in self.__KanjiNumber2Number:
                            stack2.append(r[0])
                    if r[1] == '':
                        stack2.append(self.__unit_mapper[sp])
                        break
                    else:
                        as_str = r[1]
                else:
                    as_str = r[0]
            else:
                if as_str in self.__KanjiNumber2Number:
                    stack2.append(as_str)

        number = 0
        base_unit = unit = 1
        for word in stack2:
            if word in self.__KANJI_UNIT_MAPPER.values():
                base_unit = word
                unit = 1
                continue
            elif word in self.__KANJI_MINOR_UNITS.values():
                unit = word
                continue
            else:
                number += self.__KanjiNumber2Number[word] * unit * base_unit
                unit = 1
        return number

    def kanji2number(self,
           val : str,
        )-> JpNumber:
        """ Conver Number to Kanji Number
        Parameters
        ----------
        val: str
            any number as string
        Returns:
           converted number: JpNumber
        """
        replace_comma = {
            '，': '',
            ',': '',
        }

        as_kanji = val

        replace_daiji = {
            '零': '〇',
            '壱': '一',
            '弐': '二',
            '参': '三',
            '肆': '四',
            '伍': '五',
            '陸': '六',
            '漆': '七',
            '捌': '八',
            '玖': '九',
            '拾': '十',
            '佰': '百',
            '仟': '千',
            '萬': '万',
        }
        daiji_contain = any( x in val for x in replace_daiji.keys() )
        if daiji_contain:
            for src, dst in replace_daiji.items():
                val = re.sub( src, dst, val)

        for src, dst in replace_comma.items() :
                val = re.sub(src, dst, val)

        stack = list()
        for c in val:
            if ( c in self.__unit_mapper
                 or c in self.__KanjiNumber2Number):
                 stack.append(c)

        as_str = str().join(stack)
        if as_str == '':
            return JpNumber( -1, '', as_kanji)

        number = 0

        all_unit = ( list(self.__KANJI_MINOR_UNITS.keys())
                     + list(self.__KANJI_UNIT_MAPPER.keys()) )
        unit_contain = any(x in val for x in all_unit)
        arabic_contain = any( x in val for x in self.__MULTIBYTE_DIGITS )

        if not unit_contain:
            number = self.kanji2arabic_parser(as_str)
        elif  arabic_contain :
            number = self.kanji2mix_parser(as_str)
        else:
            number = self.kanji2number_parser(as_str)

        as_str = str(number)
        return JpNumber( number, as_str, as_kanji)

    def parse_kanjinumber(self,
            text: str,
            startwith: Optional[re.Pattern]=None,
            endwith: Optional[re.Pattern]=None,
        ) -> Optional[ParseNumber]:

        if not startwith:
            startwith = f'{NOT_KANJI_ARABIC_NUMBERS}'

        if not endwith:
            endwith = f'{NOT_KANJI_ARABIC_NUMBERS}' + r'(.*)?'

        kanji_number = re.compile(
             startwith + f'{KANJI_ARABIC_NUMBERS}' + endwith,
             re.UNICODE
        )

        match =  kanji_number.search(text)

        if match:
            if len(match.groups()) == 1 :
                if isinstance(startwith, str):
                    prefix = text.split(startwith)[0] + startwith
                else:
                    prefix = ''
                if isinstance(endwith, str):
                    suffix = endwith + text.split(endwith)[1]
                else:
                    suffix = ''
                text = match.group(1)
            else:
                prefix = match.group(1)
                text = match.group(2)
                if match.group(3):
                    suffix = str().join(match.groups()[2:])
                else:
                    suffix = ''
            number = self.kanji2number(text)
            result = ParseNumber(prefix,  number, suffix)
        else:
            result = None

        return  result

    def parse_number(self,
            text: str,
            startwith: Optional[re.Pattern]=None,
            endwith: Optional[re.Pattern]=None,
        ) -> Optional[ParseNumber]:

        if not startwith:
            startwith = f'{NOT_NUMBERS}'

        if not endwith:
            endwith = f'{NOT_NUMBERS}' + r'(.*)?'

        number = re.compile(
             startwith + f'{NUMBERS}' + endwith,
             re.UNICODE
        )

        match =  number.search(text)

        if match:
            if len(match.groups()) == 1 :
                if isinstance(startwith, str):
                    prefix = text.split(startwith)[0] + startwith
                else:
                    prefix = ''
                if isinstance(endwith, str):
                    suffix = endwith + text.split(endwith)[1]
                else:
                    suffix = ''
                text = match.group(1)
            else:
                prefix = match.group(1)
                text = match.group(2)
                if match.group(3):
                    suffix = str().join(match.groups()[2:])
                else:
                    suffix = ''
            number = self.number2kanji(text)
            result = ParseNumber(prefix,  number, suffix)
        else:
            result = None

        return  result

    def normalize_kanjinumber(self,
            text: str,
            startwith: Optional[re.Pattern]=None,
            endwith: Optional[re.Pattern]=None,
            max_try: int=5
        ) -> Optional[str]:

       normalized_text =''
       do_normalize  = True
       count = max_try
       while do_normalize and count >0:
           result = self.parse_number(text, startwith, endwith)
           if result:
               if text != result.number.as_kanji:
                   kanji_number = self.number2kanji(result.number.number)
                   normalized_text  = ( str(result.prefix)
                                + str(kanji_number.as_kanji)
                                + str(result.suffix) )
                   text = normalized_text


           result = self.parse_kanjinumber(text, startwith, endwith)
           if result:
               if text != result.number.as_kanji:
                   kanji_number = self.number2kanji(result.number.number)
                   normalized_text  = ( str(result.prefix)
                                + str(kanji_number.as_kanji)
                                + str(result.suffix ) )
                   text = normalized_text

           count -= 1
           arabic_contain = any( x in text for x in self.__MULTIBYTE_DIGITS )
           digits_contain = any( str(x) in text for x in range(10))
           do_normalize = arabic_contain or digits_contain

       return text




import os
import numpy as np
import pandas as pd
import re
from typing import Optional, Union, Any, NamedTuple
from dataclasses import dataclass, field
from pathlib import Path
from .singledispatchmethod import singledispatchmethod
from .immutable_dict import ImmutableDict
from .checkdigit import calc_checkdigit, validate_checkdigit
from .jp_prefecture import JpPrefecture
from .utils import is_alpha

class Geodetic(NamedTuple):
    latitude: float
    longitude: float

@dataclass
class JpTown(object):
    prefecture: str
    city: str
    town: str
    prefCode: int=field(repr=False, default=None)
    cityCode: int=field(repr=False, default=None)
    geodetic: Geodetic=field(repr=False, default=None)

class JpCity(JpPrefecture):
    def __init__(self, enable_town:bool=False):
        # City Code: JIS X 0402

        super().__init__()
        self.cities = pd.read_csv(Path(__file__).parent / 'data/cities.csv',
                                  index_col = 0)
        self.cities['prefCode'] = pd.to_numeric(self.cities['prefCode'],
                                                downcast='integer')
        self.cities['cityCode'] = pd.to_numeric(self.cities['cityCode'],
                                                downcast='integer')
        self.cities['latitude'].astype(float)
        self.cities['longitude'].astype(float)
        self.cities['bigCityFlag'] = pd.to_numeric(self.cities['bigCityFlag'],
                                                downcast='integer')

        self.enable_town = os.environ.get('JP_PREFECTURE_ENABLE_TOWN',
                                           default=enable_town)
        if self.enable_town:
            self.towns = pd.read_csv(Path(__file__).parent / 'data/towns.csv',
                                  index_col = 0)
            self.towns['prefCode'] = pd.to_numeric(self.towns['prefCode'],
                                                downcast='integer')
            self.towns['cityCode'] = pd.to_numeric(self.towns['cityCode'],
                                                downcast='integer')
            self.towns['latitude'].astype(float)
            self.towns['longitude'].astype(float)
            self.towns['bigCityFlag'] = pd.to_numeric(self.towns['bigCityFlag'],
                                                downcast='integer')

        self.__citycode2name = ImmutableDict({
            code: name
            for code, name in zip(self.cities.cityCode,
                                  self.cities.cityName)
        })

        self.__citycode2alphabet = ImmutableDict({
            code: alphabet
            for code, alphabet in zip(self.cities.cityCode,
                                  self.cities.cityAlphabet)
        })

        self.__cityname2code = ImmutableDict({
            **{name: code
               for name, code in zip(self.cities.cityName,
                                     self.cities.cityCode)},
            **{name: code
               for name, code in zip(self.cities.cityAlphabet,
                                     self.cities.cityCode)},
            **{name.title(): code
               for name, code in zip(self.cities.cityAlphabet,
                                     self.cities.cityCode)},
        })

        self.__cityname2alphabet = ImmutableDict({
            **{name: alphabet
               for name, alphabet in zip(self.cities.cityName,
                                     self.cities.cityAlphabet)},
            **{name: alphabet
               for name, alphabet in zip(self.cities.cityAlphabet,
                                     self.cities.cityAlphabet)},
            **{name.title(): alphabet
               for name, alphabet in zip(self.cities.cityAlphabet,
                                     self.cities.cityAlphabet)},
        })

        self.__cityalphabet2name = ImmutableDict({
            **{alphabet: name
               for alphabet, name in zip(self.cities.cityName,
                                     self.cities.cityName)},
            **{alphabet: name
               for alphabet, name in zip(self.cities.cityAlphabet,
                                     self.cities.cityName)},
            **{alphabet.title(): name
               for alphabet, name in zip(self.cities.cityAlphabet,
                                     self.cities.cityName)},
        })

    @singledispatchmethod
    def cityname2code(self,
            arg: Any
        ) -> Union[Optional[Union[int,str]],list,pd.Series]:
        """ dispatch function """
        raise TypeError('Unsupport Type')

    @cityname2code.register(type(None))
    def _cityname2code_none(self,
            name: str,
            ignore_case: bool=False,
            checkdigit: bool=False,
            as_str: bool=False,
        ) -> Optional[Union[int,str]]:
        """ Catch None and return None """
        return None

    @cityname2code.register(str)
    def _cityname2code_str(self,
            name: str,
            ignore_case: bool=False,
            checkdigit: bool=False,
            as_str: bool=False,
        ) -> Optional[str]:
        """ Convert cityName to cityCode """
        try:
            name = [name, name.title()][ignore_case]
            code = self.__cityname2code[name]
            if code:
                code = [code, str(code).zfill(5)][as_str]
                code = [code, calc_checkdigit(code)][checkdigit]
        except KeyError:
            code = None
        return code

    @cityname2code.register(re.Pattern)
    def _cityname2code_str(self,
            name: re.Pattern,
            ignore_case: bool=False,
            checkdigit: bool=False,
            as_str: bool=False,
        ) -> list:
        """ Convert cityName to cityCode """
        cities = [ re.search(name, x) for x in self.__cityname2code.keys() ]
        codes = list()
        for x in cities:
            if x:
                code = self.__cityname2code[x.group(0)]
                if code not in codes:
                    codes.append(code)
        return codes

    @cityname2code.register(list)
    def _cityname2code_list(self,
            name_list: list,
            ignore_case: bool=False,
            checkdigit: bool=False,
            as_str: bool=False,
        ) -> list:
        """ Convert list of cityName to cityCode """
        code = [self.cityname2code(x, ignore_case, checkdigit)
                                   for x in name_list]
        return code

    @cityname2code.register(pd.Series)
    def _cityname2code_series(self,
            name_series: pd.Series,
            ignore_case: bool=False,
            checkdigit: bool=False,
            as_str: bool=False,
        ) -> pd.Series:
        """ Convert pandas series of cityName to cityCode """
        try:
            name_series = [name_series,
                           name_series.str.title()][ignore_case]
            code = name_series.map(self.__cityname2code)
            code = [code,
                    code.map(calc_checkdigit)][checkdigit]
        except KeyError:
            code = pd.Series([])
        return code


    @singledispatchmethod
    def citycode2name(self, arg: Any) -> Union[Optional[str],list,pd.Series]:
        """ Convert cityCode to cityName """
        raise TypeError('Unsupport Type')

    @citycode2name.register(type(None))
    def _citycode2name_none(self,
            code: None,
            ascii: bool=False,
        ):
        return None

    @citycode2name.register(int)
    def _citycode2name_int(self,
            code: int,
            ascii: bool=False,
        ) -> Optional[str]:
        """ Convert cityCode to cityName
            if set ``True`` to ascii, return cityname as alphabet_name
        """
        if len(str(code)) == 6:
            code = validate_checkdigit(code)  # type: ignore
        try:
            name = [ self.__citycode2name[code],
                     self.__citycode2alphabet[code]][ascii]
        except KeyError:
            name = None
        return name

    @citycode2name.register(str)
    def _citycode2name_str(self,
            code: str,
            ascii: bool=False,
        ) -> Optional[str]:
        """ Convert cityCode to cityName
            if set ``True`` to ascii, return cityname as alphabet_name
        """
        if len(code) == 6:
            code = validate_checkdigit(int(code), 5)  # type: ignore
        else:
            code = int(code)   # type: ignore
        try:
            name = [ self.__citycode2name[code],
                     self.__citycode2alphabet[code]][ascii]
        except:
            name = None
        return name

    @citycode2name.register(list)
    def _citycode2name_list(self,
            code_list: list,
            ascii: bool=False,
        ) -> list:
        """ Convert list of cityCode to CityName
            if set ``True`` to ascii, return cityname as alphabet_name
        """
        name = [self.citycode2name(x, ascii) for x in code_list]
        return name

    @citycode2name.register(pd.Series)
    def _citycode2name_series(self,
            code_series: pd.Series,
            ascii: bool=False,
        ) -> pd.Series:
        """ Convert pandas series of cityCode to cityName
            if set ``True`` to ascii, return cityname as alphabet_name
        """
        try:
            code_series.astype(int)
            name = [ code_series.map(self.__citycode2name),
                     code_series.map(self.__citycode2alphabet)][ascii]
        except:
            name = pd.Series([])
        return name

    @singledispatchmethod
    def cityname2normalize(self, arg: Any
        ) -> Union[Optional[int],list,pd.Series]:
        """ Normalize cityName
            if set ``True`` to ascii, return cityName as alphabet_name
        """
        raise TypeError('Unsupport Type')

    @cityname2normalize.register(type(None))
    def _cityname2normalize_none(self,
            name: None,
            ascii: bool=False,
            ignore_case: bool=False
        ):
        return None

    @cityname2normalize.register(str)
    def _cityname2normalize_str(self,
            name: str,
            ascii: bool=False,
            ignore_case: bool=False
        ) -> Optional[str]:
        """ Normalize cityName
            if set ``True`` to ascii, return cityName as alphabet_name
        """
        try:
            name = [name, name.title()][ignore_case]
            name = [ self.__cityalphabet2name[name],
                     self.__cityname2alphabet[name] ][ascii]
        except KeyError:
            name = None    # type: ignore
        return name

    @cityname2normalize.register(list)
    def _cityname2normalize_list(self,
            name_list: list,
            ascii: bool=False,
            ignore_case: bool=False
        ) -> list:
        """ Convert list of cityName to cityCode
            if set ``True`` to ascii, return cityName as alphabet_name
        """
        name = [self.cityname2normalize(x,ascii,ignore_case)
                                        for x in name_list]
        return name

    @cityname2normalize.register(pd.Series)
    def _cityname2normalize_series(self,
            name_series: pd.Series,
            checkdigit=False,
            ascii: bool=False,
            ignore_case: bool=False
        ) -> pd.Series:
        """ Convert pandas series of cityName to cityCode
            if set ``True`` to ascii, return cityName as alphabet_name
        """
        try:
            name_series = [name_series,
                           name_series.str.title()][ignore_case]
            name = [ name_series.map(self.__cityalphabet2name),
                     name_series.map(self.__cityname2alphabet) ][ascii]
        except KeyError:
            name = pd.Series([])
        return name

    def get_prefcode(self, citycode: Union[int, str])-> Optional[int]:
        if isinstance(citycode, int):
            citycode = str(citycode)
        try:
            return int(citycode[:2])
        except:
            return None

    def citycode2normalize(self,
            citycode: Union[int, str],
            as_str: bool=False,
        )-> Optional[int]:
        if isinstance(citycode, int):
            citycode = str(citycode).zfill(5)

        try:
            if len(citycode) == 6:
                citycode = validate_checkdigit(citycode)  # type: ignore
            citycode = [int(citycode), citycode][as_str]
        except:
            citycode = None

        return citycode

    def findcity(self,
            name: re.Pattern,
            ignore_case: bool=False,
            ascii: bool=False,
        ) -> list:
        flag = [0, re.IGNORECASE][ignore_case]
        result = [ re.search(name, x, flag)
                   for x in self.__cityname2code.keys() ]
        cities = list()
        for x in result:
            if x:
                city = x.group(0).title()
                if city in self.__cityname2code.keys():
                    name = [ city, self.__cityname2alphabet[city] ][ascii]
                    if name not in cities:
                        cities.append(name)

        return cities

    def findtown(self,
            name: re.Pattern,
            city: Optional[Union[str,int]]=None
        )->list:
        if not self.enable_town:
            raise NotImplementedError('Town data does not enabled.')

        if city:
            city = self.cityname2code(city)
            df = self.towns[self.towns['cityCode'] == city]
        else:
            df = self.towns.copy()
        df = df[df['town'].str.match(name)]
        df['cityName'] = df['cityCode'].map(self.__citycode2name)
        df['prefecture'] = df['cityName'].map(self.cityname2prefecture)
        df['prefCode'] = df['cityName'].map(self.cityname2prefcode)
        df = df.loc[:, [
                    'prefecture', 'prefCode',
                    'cityName', 'cityCode',
                    'town', 'latitude', 'longitude']]
        towns_list = list()
        for town in df.values:
            towns_list.append( JpTown(
                     prefecture=town[0], prefCode=town[1],
                     city=town[2], cityCode=town[3],
                     town=town[4],
                     geodetic=Geodetic(town[5], town[6])
                 )
              )
        return towns_list

    @singledispatchmethod
    def cityname2prefcode(self, arg: Any,
        ) -> Union[Optional[int],list,pd.Series]:
        """ dispatch function """
        raise TypeError('Unsupport Type')

    @cityname2prefcode.register(type(None))
    def _cityname2prefcode_none(self,
            name: None,
            ignore_case: bool=False
        ):
        """ Catch None and return None """
        return None

    @cityname2prefcode.register(str)
    def _cityname2prefcode_str(self,
            name: str,
            ignore_case: bool=False
        ) -> Optional[int]:
        """ Convert CityName to Prefecture Code"""
        name = [name, name.title()][ignore_case]
        code = self.__cityname2code[name]
        prefcode = self.get_prefcode(code)
        return prefcode

    @cityname2prefcode.register(list)
    def _cityname2prefcode_list(self,
            name_list: list,
            ignore_case: bool=False
        ) -> list:
        """ Convert list of cityName to Prefecture Code """
        code = [self.cityname2prefcode(x, ignore_case) for x in name_list]
        return code

    @cityname2prefcode.register(pd.Series)
    def _cityname2prefcode_series(self,
            name_series: pd.Series,
            ignore_case: bool=False
        ) -> pd.Series:
        """ Convert pandas series of cityName to Prefecture Code """
        try:
            name_series = [name_series,
                           name_series.str.title()][ignore_case]
            code = name_series.map(self.__cityname2code)
            code = code.map(self.get_prefcode)
        except KeyError:
            code = pd.Series([None])
        return code

    @singledispatchmethod
    def cityname2prefecture(self, arg: Any) -> Union[Optional[str],list,pd.Series]:
        """ dispatch function """
        raise TypeError('Unsupport Type')

    @cityname2prefecture.register(type(None))
    def _cityname2prefecture_none(self,
            name: None,
            ascii: bool=False,
            ignore_case: bool=False
        ):
        """ Catch None and return None """
        return None

    @cityname2prefecture.register(str)
    def _cityname2prefecture_str(self,
            name: str,
            ascii: bool=False,
            ignore_case: bool=False
        ) -> Optional[str]:
        """ Convert CityName to Prefecture Name"""
        name = [name, name.title()][ignore_case]
        code = self.cityname2prefcode(name)
        name = code and self.code2name(code, ascii)  # type: ignore
        return name

    @cityname2prefecture.register(list)
    def _cityname2prefecture_list(self,
            name_list: list,
            ascii: bool=False,
            ignore_case: bool=False
        ) -> list:
        """ Convert list of CityName to Prefecture Name"""
        name = [self.cityname2prefecture(x, ascii,ignore_case)
                                         for x in name_list]
        return name

    @cityname2prefecture.register(pd.Series)
    def _cityname2prefecture_series(self,
            name_series: pd.Series,
            ascii: bool=False,
            ignore_case: bool=False
        ) -> pd.Series:
        """ Convert pandas series of cityName to Prefecture Name """
        try:
            name_series = [name_series,
                           name_series.str.title()][ignore_case]
            code = name_series.map(self.cityname2prefcode)
            name = code.apply(self.code2name, ascii=ascii)
        except KeyError:
            # v = [None]
            name = pd.Series([None])
        # return pd.Series(v)
        return name


    @singledispatchmethod
    def cityname2geodetic(self,
            arg: Any,
        ) -> Union[Optional[tuple],list,pd.DataFrame]:
        """ dispatch function """
        raise TypeError('Unsupport Type')

    @cityname2geodetic.register(str)
    def _cityname2geodetic_str(self,
            name: str,
            ignore_case: bool=False
        ) -> Optional[tuple]:
        """ Return Latitude and Longitude of CityName """
        code = self.cityname2code(name, ignore_case)
        if code:
            city = self.cities.loc[self.cities['cityCode'] == code]
            geodetic = Geodetic( city['latitude'].values[0],
                         city['longitude'].values[0] )
        else:
            geodetic = None    # type: ignore
        return geodetic

    @cityname2geodetic.register(list)
    def _cityname2geodetic_list(self,
            name_list: list,
            ignore_case: bool=False
        ) -> list:
        """ Return Latitude and Longitude of CityName """
        code = self.cityname2code(name_list, ignore_case)
        geodetic = list()
        for c in code:
            city = self.cities.loc[self.cities['cityCode'] == c]
            pos = Geodetic( city['latitude'].values[0],
                    city['longitude'].values[0] )
            geodetic.append(pos)
        if len(geodetic) == 0:
            geodetic = None    # type: ignore
        return geodetic

    @cityname2geodetic.register(pd.Series)
    def _cityname2geodetic_series(self,
            name_series: pd.Series,
            ignore_case: bool=False
        ) -> pd.DataFrame:
        """ Return Latitude and Longitude of CityName """
        code = name_series.apply(self.cityname2code, ignore_case)
        geodetic = self.cities.loc[self.cities['cityCode'].isin(code.values),
                                   ['latitude','longitude']]
        geodetic.insert(0, 'cityName', name_series.values )
        return geodetic

    @singledispatchmethod
    def citycode2geodetic(self,
            arg: Any,
        ) -> Union[Optional[tuple],list,pd.DataFrame]:
        """ dispatch function """
        raise TypeError('Unsupport Type')

    @citycode2geodetic.register(str)
    @citycode2geodetic.register(int)
    def _citycode2geodetic(self,
            code: Union[int,str],
            ignore_case: bool=False
        ) -> Optional[tuple]:
        """ Return Latitude and Longitude of CityName """

        citycode = self.citycode2normalize(code)
        if citycode:
            city = self.cities.loc[self.cities['cityCode'] == citycode]
            geodetic = Geodetic( city['latitude'].values[0],
                         city['longitude'].values[0] )
        else:
            geodetic = None    # type: ignore
        return geodetic

    @citycode2geodetic.register(list)
    def _citycode2geodetic_list(self,
            code_list: list,
            ignore_case: bool=False
        ) -> list:
        """ Return Latitude and Longitude of CityName """
        city_codes = [ self.citycode2normalize(x) for x in code_list ]
        geodetic = list()
        try:
            for c in city_codes:
                city = self.cities.loc[self.cities['cityCode'] == c]
                pos = Geodetic( city['latitude'].values[0],
                        city['longitude'].values[0] )
                geodetic.append(pos)
        except:
            geodetic = list()

        if len(geodetic) == 0:
            geodetic = None    # type: ignore
        return geodetic

    @citycode2geodetic.register(pd.Series)
    def _citycodegeodetic_series(self,
            code_series: pd.Series,
            ignore_case: bool=False
        ) -> pd.DataFrame:
        """ Return Latitude and Longitude of CityName """
        city_codes =  code_series.map(self.citycode2normalize)
        geodetic = self.cities.loc[( self.cities['cityCode']
                                         .isin(city_codes.values)),
                                     ['cityCode','latitude','longitude']]
        return geodetic

    @singledispatchmethod
    def validate_city(self, arg: Any) -> Optional[Union[bool,list,pd.Series]]:
        """ dispatch function """
        raise TypeError('Unsupport Type')

    @validate_city.register(type(None))
    def _validate_city_none(self,
            name: None,
            ignore_case: bool=False
        ) -> bool:
        """ Catch None and return None """
        return False

    @validate_city.register(str)
    def _validate_city_str(self,
            name: str,
            ignore_case: bool=False
        ) -> bool:
        """ validate_city a cityName """
        try:
            name = [name, name.title()][ignore_case]
            v = name in self.__cityname2code.keys()
        except:
            v = False
        return v

    @validate_city.register(list)
    def _validate_city_list(self,
            name_list: list,
            ignore_case: bool=False
        ) -> list:
        """ validate_city list of cityName """
        try:
            v = [ self.validate_city(x,ignore_case) for x in name_list]
        except:
            v = [False]
        return v

    @validate_city.register(pd.Series)
    def _validate_city_series(self,
            name_series: pd.Series,
            ignore_case: bool=False
        ) -> pd.Series:
        """ validate_city pandas series of cityName """
        try:
            v = [ self.validate_city(x,ignore_case)
                                     for x in np.asarray(name_series)]
        except KeyError:
            v = [False]
        return pd.Series(v)

jp_cities = JpCity()



import re
from dataclasses import dataclass, InitVar, field
from typing import Optional
from .jp_cities import jp_cities as jp
from .jp_cities import Geodetic
from .jp_numbers import JpNumberParser

class JpAddressError(BaseException):
    pass

@dataclass
class JpAddress(object):
    zipCode: str
    prefecture: str
    city: str
    street: str
    prefCode: int=field(init=False, repr=False, default=None)
    cityCode: int=field(init=False, repr=False, default=None)
    geodetic: Geodetic=field(init=True, repr=False, default=None)

    def __post_init__(self):
        zip_parser = JpZipCode()
        self.zipCode = zip_parser.zip2normalize(self.zipCode)
        if self.city:
            self.cityCode = jp.cityname2code(self.city)
            self.city = jp.cityname2normalize(self.city)
        else:
            self.city = None

        if not self.prefecture:
            self.prefecture = jp.cityname2prefecture(self.city)
        else:
            self.prefecture = jp.name2normalize(self.prefecture)
        self.prefCode = jp.name2code(self.prefecture)
        if not self.geodetic:
            self.geodetic = jp.cityname2geodetic(self.city)

    def __str__(self):
        if self.zipCode:
            zip= f'〒{self.zipCode[:3]}-{self.zipCode[3:]} '
        else:
            zip = ''
        address = (
            f'{zip}'
            f'{self.prefecture}'
            f'{self.city}'
            f'{self.street}'
        )
        return address

class JpZipCode(object):
    __zip_pattern = (
       r'(〒)? *(\d{3}-\d{4}|\d{7})? *'
    )
    def __init__(self):
        self.zip_re = re.compile(self.__zip_pattern, re.UNICODE)

    def zip2normalize(self,
            zipcode: Optional[str]=None
        ) -> Optional[str]:
        if not zipcode:
            return None

        r = self.zip_re.search(zipcode)
        result = None
        if r:
            symbol = r.group(1)
            code = r.group(2)
            if not code:
                return None
            result = code.replace('-','')
        return result

class JpAddressParser(JpZipCode):
    __address_pattern = (
        r'(〒? *\d{3}-\d{4}|〒? *\d{7})? *'
        r'(?P<Prefecture>...?[都道府県]|..?)? *'
        r'('
          r'(?:'
          r'旭川|伊達|石狩|盛岡|奥州|田村|南相馬|那須塩原|'
          r'東村山|武蔵村山|羽村|十日町|上越|富山|野々市|大町|'
          r'蒲郡|四日市|姫路|大和郡山|廿日市|下松|岩国|田川|'
          r'大村|宮古|富良野|別府|佐伯|黒部|小諸|塩尻|玉野|周南'
          r')市 *|'
        r'(?:余市|高市|[^市]{2,3}?)'
        r'郡(?:玉村|大町|.{1,5}?)[町村]|'
        r'(?:.{1,4}市)?[^町]{1,4}?区|'
        r'.{1,7}?[市町村])(.+)'
    )

    def __init__(self):
        self.address_re = re.compile(self.__address_pattern, re.UNICODE)
        self.jp_number = JpNumberParser()
        super().__init__()

    def parse_address(self, address):
        r = self.address_re.search(address)
        result = None
        if r:
            prefecture = jp.name2normalize(r.group('Prefecture'))
            if not prefecture:
                city = r.group('Prefecture') + r.group(3)
            else:
                city = r.group(3).replace(' ', '')
                city = city.replace('　', '') # Kanji Space
                city_normal = jp.cityname2normalize(city)
                if not city_normal:
                    city = jp.findcity(r.group(2) + '.*' + city)[0]
                else:
                    city = city_normal
            street = re.sub('^[ 　]*', '', r.group(4))
            normalized_street = self.jp_number.normalize_kanjinumber(street)
            town = normalized_street.split('丁目')
            if len(town) >1:
                town = town[0] + '丁目'
            else:
                town = re.split('[-ー]', normalized_street)
                if len(town) >1:
                    town = town[0] + '丁目'
                else:
                    town = normalized_street[:6]
            try:
                town = jp.findtown(town, city)[0]
                geodetic = town.geodetic
            except:
                geodetic = None

            result = JpAddress(zipCode=r.group(1),
                               prefecture=prefecture, city=city,
                               street=street,
                               geodetic=geodetic,
                               )
        return result

